/**
 * @fileOverview A communications server that implements real-time,
 * bi-directional network connectivity using HTTP and provide basic HTTP
 * Server functionality for MUD Web Client support
 * @author Dino M. Gambone
 */

"use strict";

//System and File I/O modules
var System = require("../System");
var Util = System.Util;
var FileSystem = System.FileSystem;
var Path = System.Path;
var EventEmitter = new require("events").EventEmitter;

//Networking modules
var URL = System.URL;
var HTTP = System.HTTP;
var SocketIO = System.SocketIO;

//Game Engine Modules
var HTTPClient = require("./Client");

//Load in known mime-types
/**
 * Supported MIME types
 * @memberOf HTTPServer.Server
 * @internal
 * @private
 */
var MIMETypes = {};
FileSystem.readFile(Path.resolve(__dirname, "mime-types.json"), function(err, data)
{
	if(err)
		throw err;

	MIMETypes = JSON.parse(data);
});

/**
 * The default Server Settings
 * @memberOf HTTPServer
 */
var DefaultSettings =
{
	/**
	 * The port to list for HTTP connections
	 */
	port: 80

	/**
	 * The local file path to treat as the base path for the web files
	 */
	, basePath: "."

	/**
	 * The default file to serve is none is requested
	 */
	, defaultFile: "index.html"

	/**
	 * HTTP Methods supported by the server
	 */
	, methods: ["GET", "POST"]
};

/**
  * Game Server that uses the HTTP protocol for communications
  * @extends EventEmitter
  * @class
  * @name HTTPServer.Server
  */
var HTTPServer = Object.create(EventEmitter.prototype, Util.toPropertiesObject(
/** @lends HTTPServer.Server# */
{
	//Event Documentation
	/**
	 * Raised when the server outputs log information
	 * @event
	 * @memberOf HTTPServer.Server#
	 * @name log
	 * @param {String} msg The log message
	 */

	/**
	 * Raised when a client connects to the server
	 * @event
	 * @memberOf HTTPServer.Server#
	 * @name clientConnected
	 * @param {Object} client The client connection object
	 */

	/**
	 * The list of connected clients.  This member should only be used
	 * internally.
	 * @field
	 * @private
	 */
	_clients: {}

	/**
	 * The server settings
	 * @type {Object}
	 */
	, settings: {}

	/**
	  * Sends a resource to the requesting client
	  * @private
	  */
	, _sendResource: function(resourcePath, response)
	{
		//Read in the file and send back the data
		FileSystem.readFile(resourcePath, function(err, data)
		{
			//Uh-oh, something went wrong and we don't know what it was
			if(err)
			{
				response.writeHead(500, "Server encountered an error");
				response.write(err.toString());
				response.end();
				return;
			}

			//Get the mime type
			var mime_type = MIMETypes[Path.extname(resourcePath)];
			if(mime_type !== undefined)
				response.setHeader("Content-Type", mime_type);

			response.writeHead(200, "OK");
			response.write(data);
			response.end();
		});
	}

	/**
	  * Handles a single HTTP request
	  *	@param {Object} request The HTTP request object
	  *	@param {Object} response The HTTP response object
	  * @private
	  */
	, _onHTTPRequest: function(request, response)
	{
		var self = this;
		var http_methods = this.settings.methods;

		//If the HTTP method isn't an HTTP method supported
		//then return the standard 405 error with methods
		//supported
		if(http_methods.indexOf(request.method.toUpperCase()) == -1)
		{
			response.setHeader("Allow", http_methods.join(","));
			response.writeHead(405, "Method not allowed");
			return;
		}

		//Get the path name of the file requested
		var path = URL.parse(request.url).pathname;

		//If the path is the root directory, default to the
		//default resource specified in the options
		if(path == "/")
			path = this.settings.defaultFile;

		//Resolve the path to be relative to this executing file
		path = Path.join(this.settings.basePath, path);

		//Check to see if the file being requested exists
		Path.exists(path, function(fileExists)
		{
			//The file doesn't exist, send back a standard 404 error.
			if(!fileExists)
			{
				self.log("File not found: " + path);
				response.writeHead(404, "File not found");
				response.end();
				return;
			}

			//Everything looks ok, send the resource to the client
			self._sendResource(path, response);
		});
	}

	/**
	  * Callback function that handles when a client connects to the server
	  * @private
	  */
	, _onClientConnect: function(socket)
	{
		//Create a new client
		var client = HTTPClient.create(this, socket);
		this.emit("clientConnect", client);

		//If after notifying everyone the client is still connected,
		//let it know that it is now connected
		if(client.isConnected)
		{
			this.addClient(client);
			client.send("connected", {});
			this.emit("clientConnected", client);
		}
	}

	/**
	 *
	 * Starts listening on a port for incoming connections
	 * @param {Object} settings Settings for the server
	 */
	, start: function(settings)
	{
		var self = this;

		//Incorporate the settings into the default settings
		var keys = Object.getOwnPropertyNames(settings);
		var key = null;
		while(keys.length)
		{
			key = keys.shift();
			this.settings[key] = settings[key];
		}

		//Create the actual HTTP server with a callback to
		//the onHTTPRequest function
		var server = HTTP.createServer();
		server.on("request", function(request, response)
			{
				self._onHTTPRequest(request, response);
			});

		//Create the Socket.IO object and assign it to the HTTP server
		//with a callback to the onClientConnect function
		var io = SocketIO.listen(server,
			{
				log: function(message)
				{
						self.log(message);
				}
			});
		io.set("log level", 0);

		io.sockets.on("connection", function(client)
			{
				self._onClientConnect(client);
			});

		//Remeber the objects
		Object.defineProperty(this, "HTTPServer", {value: server, enumerable: false});
		Object.defineProperty(this, "SocketIO", {value: io, enumerable: false});

		self.log("Listening on port " + this.settings.port);

		//Start listening for connections
		server.listen(this.settings.port);
	}

	/**
	  * Writes a message to a log
	  * @param {String} message The message to write to the log file
	  * @private
	  */
	, log: function(message)
	{
		this.emit("log", message);
	}

	/**
	  * Returns a client object based on a unique ID
	  * @param {string} clientID The using Client ID
	  */
	, getClient: function(clientID)
	{
		var client = this._clients[clientID];

		if(!client)
			return null;

		return client;
	}

	/**
	  * Adds a client to the list of clients managed by this server
	  * @param {Object} client The client object to remove
	  */
	, addClient: function(client)
	{
		if(client == null)
			throw new TypeError();

		//First make sure this client isn't already in the list
		if(this.getClient(client.ID) != null)
			return;

		this._clients[client.ID] = client;
	}

	/**
	  * Removes a client to the list of clients managed by this server
	  * @param {Object} client The client object to remove
	  */
	, removeClient: function(client)
	{
		if(!client || !client.ID)
			return;

		var client_id = client.ID;
		delete this._clients[client_id]
	}

	/**
	  * Sends a message to all clients
	  * @param {String} message The message to send to all the clients
	  */
	, broadcast: function(message)
	{
		var client_id = "";
		var client;

		for(client_id in this._clients)
		{
			client = this._clients[client_id];
			client.send(message);
		}
	}
}));

module.exports =
{
	create: function()
	{

		var settings = Object.create(DefaultSettings);
		return Object.create(HTTPServer, Util.toPropertiesObject(
			{
				settings: settings
				, _clients: {}
			}));
	}
}
