var Util = require("./System").Util;
var EventEmitter = new require("events").EventEmitter;

/**
  * A message queue that receives messages and automatically pops them off
  * as quickly as possible in sequential order.
  * @extends EventEmitter
  * @class
  * @name MessagePump
 */
var MessagePump = Object.create(EventEmitter.prototype, Util.toPropertiesObject(
	/** @lends MessagePump# */
{
	//Event Documentation
	/**
	 * Raised when a message is pushed onto the queue.  Delayed messages do
	 * not trigger this event until they are actually pushed onto the queue.
	 * @memberOf MessagePump#
	 * @name push
	 * @param {Message} message The message that was pushed onto the queue
	 * @event
	 */
	
	/**
	 * Raised when a message is popped from the message queue.
	 * @memberOf MessagePump#
	 * @name message
	 * @param {Message} message The message that was popped
	 * @event
	 */
		
	/**
	 * An array of all messages pending processing
	 * @private
	 */
	_queue: []
	
	/**
	 * An array of all delayed messages pending processing
	 * @private
	 */
	, _delayedQueue: []

	/**
	 * Pushes a new message onto the queue
	 * @function
	 * @param {DelayedMessage} message The message to push on the queye
	 * @param {number} delay The delay, in milliseconds, that the message should
	 * wait before being pushed.  The message will be processed, roughly, at
	 * that time, ahead of any other messages.
	 */
	, pushDelayed: function(message, delay)
	{
		var self = this;
		
		//Push the message onto the delayedQueue after the timeout
		setTimeout(function()
			{
				self._delayedQueue.push(message);
	
				//Notify anyone who cares that there is a new message
				self.emit("push", message, delay);
	
				//Trigger another call to processNextMessage on the next loop
				if(self._delayedQueue.length === 1)
					self.scheduleNextTick();
			}, delay);
	}

	/**
	 * Pushes a new message onto the queue
	 * @function
	 * @param {Message} message The message to push on the queuye
	 */
	, push: function(message)
	{
		//Push the message onto the queue
		this._queue.push(message);
		
		//Notify anyone who cares that there is a new message
		this.emit("push", message);

		//Trigger another call to processNextMessage on the next loop
		if(this._queue.length === 1)
			this.scheduleNextTick();
	}
	
	/**
	 * Processes the next pending message on the queue
	 * @function
	 */
	, processNextMessage: function()
	{
		var message = null;

		//Get the next message from the delayed message queue
		if(this._delayedQueue.length)
			message = this._delayedQueue.shift();

		//Get the next message from the normal message queue if there
		//isn't one from the delayed queue.
		if(!message && this._queue.length)
			message = this._queue.shift();			

		//If there is nothing to process, then wait until something is
		//pushed onto the queue
		if(!message)
			return;
		
		//Process the next message
		this.emit("message", message);

		//Trigger another call to processNextMessage on the next loop
		this.scheduleNextTick();
	}

	/**
	 *	Schedules a message to be processed on the next tick, if needed
	 *	@private
	 */
	, scheduleNextTick: function()
	{
		//Trigger another call to processNextMessage on the next loop
		var self = this;
		if(this._queue.length || this._delayedQueue.length)
			process.nextTick(function() { self.processNextMessage(); });
	}
}));

module.exports = MessagePump;